/*****************************************************************************
 * libdvdcss.c: DVD reading library.
 *****************************************************************************
 * Copyright (C) 1998-2001 VideoLAN
 * $Id: input_dvd.c,v 1.68 2001/06/07 22:25:42 sam Exp $
 *
 * Authors: Stphane Borel <stef@via.ecp.fr>
 *          Samuel Hocevar <sam@zoy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include "defs.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#if defined( WIN32 )
#   include <io.h>
#   include "input_iovec.h"
#else
#   include <sys/uio.h>                                      /* struct iovec */
#endif

#include "config.h"
#include "common.h"

#include "dvdcss.h"
#include "libdvdcss.h"

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int _dvd_open  ( DVD_Handle, char *psz_target );
static int _dvd_close ( DVD_Handle );
static int _dvd_seek  ( DVD_Handle, int i_blocks );
static int _dvd_read  ( DVD_Handle, void *p_buffer, int i_blocks );
static int _dvd_readv ( DVD_Handle, void *p_iovec, int i_blocks );

/*****************************************************************************
 * DVD_Init: initialize libdvdcss
 *****************************************************************************/
extern DVD_Handle DVD_Init ( int i_flags )
{
    DVD_Handle dvdcss;

    /* Allocate the library structure */
    dvdcss = malloc( sizeof( struct dvdcss_s ) );
    if( dvdcss == NULL )
    {
        if( ! (i_flags & DVD_INIT_QUIET) )
        {
            DVD_ERROR( "could not initialize library" );
        }

        return NULL;
    }

    /* Initialize structure */
    dvdcss->i_status = DVD_STATUS_NONE;

    dvdcss->b_debug = i_flags & DVD_INIT_DEBUG;
    dvdcss->b_errors = !(i_flags & DVD_INIT_QUIET);
    dvdcss->psz_error = "no error";

    /* XXX: additional initialization stuff might come here */

    dvdcss->i_status |= DVD_STATUS_INIT;

    return dvdcss;
}

/*****************************************************************************
 * DVD_Error: return the last libdvdcss error message
 *****************************************************************************/
extern char * DVD_Error ( DVD_Handle dvdcss )
{
    return dvdcss->psz_error;
}

/*****************************************************************************
 * DVD_Open: open a DVD device, crack CSS key if disc is encrypted
 *****************************************************************************/
extern int DVD_Open ( DVD_Handle dvdcss, char *psz_target )
{
    int i_ret;

    if( ! (dvdcss->i_status & DVD_STATUS_INIT) )
    {
        _dvd_error( dvdcss, "library not initialized" );
        return -1;
    }

    if( dvdcss->i_status & DVD_STATUS_OPEN )
    {
        _dvd_error( dvdcss, "a device is already opened" );
        return -1;
    }

    i_ret = _dvd_open( dvdcss, psz_target );
    if( i_ret < 0 )
    {
        return i_ret;
    }

    i_ret = CSSTest( dvdcss );
    if( i_ret < 0 )
    {
        _dvd_error( dvdcss, "css test failed" );
        _dvd_close( dvdcss );
        return i_ret;
    }

    dvdcss->b_encrypted = i_ret;

    /* If drive is encrypted, crack its key */
    if( dvdcss->b_encrypted )
    {
        i_ret = CSSInit( dvdcss );

        if( i_ret < 0 )
        {
            _dvd_close( dvdcss );
            return i_ret;
        }
    }

    dvdcss->i_status |= DVD_STATUS_OPEN;

    return 0;
}

/*****************************************************************************
 * DVD_Seek: seek into the device
 *****************************************************************************/
extern int DVD_Seek ( DVD_Handle dvdcss, int i_blocks )
{
    if( ! (dvdcss->i_status & DVD_STATUS_OPEN) )
    {
        _dvd_error( dvdcss, "no device opened" );
        return -1;
    }

    return _dvd_seek( dvdcss, i_blocks );
}

/*****************************************************************************
 * DVD_Crack: crack the current title key
 *****************************************************************************/
extern int DVD_Crack ( DVD_Handle dvdcss, int i_title, int i_block )
{
    int i_ret;

    if( ! (dvdcss->i_status & DVD_STATUS_OPEN) )
    {
        _dvd_error( dvdcss, "no device opened" );
        return -1;
    }

    if( ! dvdcss->b_encrypted )
    {
        return 0;
    }

    /* Crack CSS title key for current VTS */
    dvdcss->css.i_title = i_title;
    dvdcss->css.i_title_pos = i_block;

    i_ret = CSSGetKey( dvdcss );

    if( i_ret < 0 )
    {
        _dvd_error( dvdcss, "fatal error in vts css key" );
        return i_ret;
    }
    else if( i_ret > 0 )
    {
        _dvd_error( dvdcss, "decryption unavailable" );
        return -1;
    }

    return 0;
}

/*****************************************************************************
 * DVD_Read: read data from the device, decrypt if requested
 *****************************************************************************/
extern int DVD_Read ( DVD_Handle dvdcss, void *p_buffer,
                                            int i_blocks,
                                            int i_flags )
{
    int i_ret;

    if( ! (dvdcss->i_status & DVD_STATUS_OPEN) )
    {
        _dvd_error( dvdcss, "no device opened" );
        return -1;
    }

    i_ret = _dvd_read( dvdcss, p_buffer, i_blocks );

    if( i_ret != i_blocks )
    {
        return i_ret;
    }

    if( dvdcss->b_encrypted && i_flags & DVD_READ_DECRYPT )
    {
        int i_index = i_blocks;

        while( i_index )
        {
            CSSDescrambleSector( dvdcss->css.pi_title_key, p_buffer );
            ((u8*)p_buffer)[0x14] &= 0x8f;
            (u8*)p_buffer += 2048;
            i_index--;
        }
    }

    return i_blocks;
}

/*****************************************************************************
 * DVD_ReadV: read data to an iovec structure, decrypt if reaquested
 *****************************************************************************/
extern int DVD_ReadV ( DVD_Handle dvdcss, void *p_iovec,
                                          int i_blocks,
                                          int i_flags )
{
    int i_ret;

    if( ! (dvdcss->i_status & DVD_STATUS_OPEN) )
    {
        _dvd_error( dvdcss, "no device opened" );
        return -1;
    }

    i_ret = _dvd_readv( dvdcss, p_iovec, i_blocks );

    if( i_ret != i_blocks )
    {
        return i_ret;
    }

    /* XXX: decrypt */

    return i_ret;
}

/*****************************************************************************
 * DVD_Close: close the DVD device
 *****************************************************************************/
extern int DVD_Close ( DVD_Handle dvdcss )
{
    int i_ret;

    if( ! (dvdcss->i_status & DVD_STATUS_OPEN) )
    {
        _dvd_error( dvdcss, "no device opened" );
        return -1;
    }

    i_ret = _dvd_close( dvdcss );

    if( i_ret < 0 )
    {
        return i_ret;
    }

    dvdcss->i_status &= ~DVD_STATUS_OPEN;

    return 0;
}

/*****************************************************************************
 * DVD_End: clean up the library
 *****************************************************************************/
extern int DVD_End ( DVD_Handle dvdcss )
{
    if( dvdcss->i_status & DVD_STATUS_OPEN )
    {
        _dvd_error( dvdcss, "a device is still open" );
        return -1;
    }

    free( dvdcss );

    return 0;
}

/* Following functions are local */

static int _dvd_open ( DVD_Handle dvdcss, char *psz_target )
{
#if defined( WIN32 )
    snprintf( buf, 7, "\\\\.\\%c:", psz_target[0] );
    (HANDLE) dvdcss->i_fd =
            CreateFile( psz_target, GENERIC_READ | GENERIC_WRITE,
                                    FILE_SHARE_READ | FILE_SHARE_WRITE,
                                    NULL, OPEN_EXISTING, 0, NULL );
    if( (HANDLE) dvdcss->i_fd == INVALID_HANDLE_VALUE )
    {
        _dvd_error( dvdcss, "failed opening device" );
        return -1;
    }

#else
    dvdcss->i_fd = open( psz_target, 0 );

    if( dvdcss->i_fd == -1 )
    {
        _dvd_error( dvdcss, "failed opening device" );
        return -1;
    }

#endif

    return 0;
}

static int _dvd_close ( DVD_Handle dvdcss )
{
#if defined( WIN32 )
    CloseHandle( (HANDLE) dvdcss->i_fd );
#else
    close( dvdcss->i_fd );
#endif

    return 0;
}

static int _dvd_seek ( DVD_Handle dvdcss, int i_blocks )
{
    off_t i_read;

#if defined( WIN32 )
    i_read = SetFilePointer( (HANDLE) dvdcss->i_fd,
                             (off_t)i_blocks * (off_t)DVD_BLOCK_SIZE,
                             NULL, FILE_BEGIN );
#else
    i_read = lseek( dvdcss->i_fd,
                    (off_t)i_blocks * (off_t)DVD_BLOCK_SIZE, SEEK_SET );

#endif

    return i_read / DVD_BLOCK_SIZE;
}

static int _dvd_read ( DVD_Handle dvdcss, void *p_buffer, int i_blocks )
{
#if defined( WIN32 )
    DWORD i_read;
    if( ReadFile( (HANDLE) dvdcss->i_fd, p_buffer,
                  (off_t)i_blocks * (off_t)DVD_BLOCK_SIZE,
                  &i_read, NULL ) == -1 )
    {
        return 0;
    }

#else
    off_t i_read;
    i_read = read( dvdcss->i_fd, p_buffer,
                   (off_t)i_blocks * (off_t)DVD_BLOCK_SIZE );

#endif

    return i_read / DVD_BLOCK_SIZE;
}

static int _dvd_readv ( DVD_Handle dvdcss, void *p_iovec, int i_blocks )
{
    off_t i_read;
    i_read = readv( dvdcss->i_fd, p_iovec, i_blocks );

    return i_read / DVD_BLOCK_SIZE;
}

